from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from sql_cli.workflow_directory_parser import WorkflowFile, get_workflow_files

if TYPE_CHECKING:
    from astro.sql.operators.transform import TransformOperator  # pragma: no cover


def to_task_list(workflow_files: list[WorkflowFile]) -> list[TransformOperator]:
    """
    Converts the list of SQL Files into a list of TranformOperator tasks
    that all have proper dependencies set.
    :param workflow_files: The list of SQL files with necessary metadata for us to
        generate tasks with dependencies
    """
    param_dict = {s.name: s.to_operator() for s in workflow_files}
    for s in workflow_files:
        for p in s.get_parameters():
            if not param_dict.get(p):
                raise ValueError(f"variable '{p}' is undefined in file {s.name}")
            param_dict[s.name].parameters[p] = param_dict[p].output
            param_dict[s.name].set_upstream(param_dict[p])
    return list(param_dict.values())


def render_tasks(directory: Path) -> list[TransformOperator]:
    """
    render_dag allows a user to take any directory and turn it into a runnable
    Airflow DAG. This function will read all SQL files, and set dependencies based
    on jinja template-based variables.

    :param directory: Base directory for SQL files. We will recursively parse
        subdirectories as well.
    :param workflow_name: the name of the Workflow you would like to run
    :param start_date: (Optional) the start date you would like to set for your run.
        defaults to 2020-01-01
    :return: a DAG that can be run by any airflow
    """
    workflow_files: list[WorkflowFile] = sorted(get_workflow_files(directory, target_directory=None))
    return to_task_list(workflow_files)



from __future__ import annotations

from pathlib import Path
from typing import Any

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader
from jinja2.meta import find_undeclared_variables
from jinja2.runtime import StrictUndefined


def find_template_variables(file_path: Path) -> set[str]:
    """
    Find template variables in given file path which needed to be declared when rendering.

    :param file_path: The file path to check for variables.

    :returns: all undeclared variables.
    """
    env = Environment(
        loader=FileSystemLoader(file_path.parent),
        undefined=StrictUndefined,
        autoescape=True,
    )
    template_source = env.loader.get_source(env, file_path.name)
    parsed_content = env.parse(template_source)
    return find_undeclared_variables(parsed_content)  # type: ignore


def render(
    template_file: Path,
    context: dict[str, Any],
    output_file: Path,
    searchpath: Path = Path(__file__).parent.parent,
) -> None:
    """
    Render the template_file with context to a file.

    :param template_file: The path to the template file.
    :param context: The context to pass to the template file.
    :param output_file: The file to output the rendered version.
    :param searchpath: The default is the project directory.
    """
    (
        Environment(
            loader=FileSystemLoader(searchpath),
            undefined=StrictUndefined,
            autoescape=True,
            keep_trailing_newline=True,
        )
        .get_template(template_file.as_posix())
        .stream(**context)
        .dump(output_file.as_posix())
    )



class DagCycle(Exception):
    """An exception raised when DAG contains a cycle."""


class InvalidProject(Exception):
    """An exception raised when Project is invalid."""


class EmptyDag(Exception):
    """An exception raised when there are no SQL files within the DAG."""


class WorkflowFilesDirectoryNotFound(Exception):
    """An exception raised when the workflow files directory does not exist."""


class ConnectionFailed(Exception):
    """An exception raised when the sql file's connection cannot be established."""

    def __init__(self, *args: object, conn_id: str) -> None:
        self.conn_id = conn_id
        super().__init__(*args)



import importlib.metadata
import os

import typer.rich_utils

from sql_cli.utils.rich import rich_format_error

# We monkey-patch rich_format_error to make it environment aware
typer.rich_utils.rich_format_error = rich_format_error

# TODO: Remove after the `astro-sdk-python` package 1.3 is released
os.environ["AIRFLOW__CORE__ENABLE_XCOM_PICKLING"] = "True"
__version__ = importlib.metadata.version("astro-sql-cli")



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import typer
from dotenv import load_dotenv
from typer import Exit

import sql_cli
from sql_cli.astro.command import AstroCommand
from sql_cli.astro.group import AstroGroup
from sql_cli.constants import DEFAULT_AIRFLOW_HOME, DEFAULT_DAGS_FOLDER
from sql_cli.exceptions import ConnectionFailed, DagCycle, EmptyDag, WorkflowFilesDirectoryNotFound
from sql_cli.utils.rich import rprint

if TYPE_CHECKING:
    from sql_cli.project import Project  # pragma: no cover

load_dotenv()
app = typer.Typer(
    name="flow",
    cls=AstroGroup,
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="rich",
)

airflow_logger = logging.getLogger("airflow")
airflow_logger.setLevel(logging.CRITICAL)
airflow_logger.propagate = False


@app.command(
    cls=AstroCommand,
    help="Print the SQL CLI version.",
)
def version() -> None:
    rprint("Astro SQL CLI", sql_cli.__version__)


@app.command(
    cls=AstroCommand,
    help="Print additional information about the project.",
)
def about() -> None:
    rprint("Find out more: https://docs.astronomer.io/astro/cli/sql-cli")


@app.command(
    cls=AstroCommand,
    help="""
    Gets the value of the configuration key set in the configuration file for the given environment.
    """,
)
def config(
    project_dir: Path = typer.Option(
        None, dir_okay=True, metavar="PATH", help="(Optional) Default: current directory.", show_default=False
    ),
    env: str = typer.Option(
        default="default",
        help="(Optional) Environment used to fetch the configuration key from.",
    ),
    key: str = typer.Option(
        default=None,
        help="(Optional) Key from the configuration whose value needs to be fetched.",
    ),
) -> None:
    from sql_cli.configuration import Config

    project_dir_absolute = project_dir.resolve() if project_dir else Path.cwd()
    project_config = Config(environment=env, project_dir=project_dir_absolute).from_yaml_to_config()
    print(getattr(project_config, key))


@app.command(
    cls=AstroCommand,
    help="Generate the Airflow DAG from a directory of SQL files.",
)
def generate(
    workflow_name: str = typer.Argument(
        default=...,
        show_default=False,
        help="name of the workflow directory within workflows directory.",
    ),
    env: str = typer.Option(
        default="default",
        help="environment to run in",
    ),
    project_dir: Path = typer.Option(
        None, dir_okay=True, metavar="PATH", help="(Optional) Default: current directory.", show_default=False
    ),
    generate_tasks: bool = typer.Option(
        default=True,
        help="whether to explicitly generate the tasks in your SQL CLI DAG",
        show_default=True,
    ),
) -> None:
    from sql_cli.project import Project
    from sql_cli.utils.airflow import retrieve_airflow_database_conn_from_config, set_airflow_database_conn

    project_dir_absolute = project_dir.resolve() if project_dir else Path.cwd()
    project = Project(project_dir_absolute)
    project.load_config(env)

    # Since we are using the Airflow ORM to interact with connections, we need to tell Airflow to use our airflow.db
    # The usual route is to set $AIRFLOW_HOME before Airflow is imported. However, in the context of the SQL CLI, we
    # decide this during runtime, depending on the project path and SQL CLI configuration.
    airflow_meta_conn = retrieve_airflow_database_conn_from_config(project.directory / project.airflow_home)
    set_airflow_database_conn(airflow_meta_conn)

    rprint(
        f"\nGenerating the DAG file from workflow [bold blue]{workflow_name}[/bold blue]"
        f" for [bold]{env}[/bold] environment..\n"
    )
    dag_file = _generate_dag(project=project, workflow_name=workflow_name, generate_tasks=generate_tasks)
    rprint("The DAG file", dag_file.resolve(), "has been successfully generated. 🎉")


@app.command(
    cls=AstroCommand,
    help="""
    Validate Airflow connection(s) provided in the configuration file for the given environment.
    """,
)
def validate(
    project_dir: Path = typer.Argument(
        None, dir_okay=True, metavar="PATH", help="(Optional) Default: current directory.", show_default=False
    ),
    env: str = typer.Option(
        default="default",
        help="(Optional) Environment used to declare the connections to be validated",
    ),
    connection: str = typer.Option(
        default=None,
        help="(Optional) Identifier of the connection to be validated. By default checks all the env connections.",
    ),
) -> None:
    from sql_cli.connections import validate_connections
    from sql_cli.project import Project
    from sql_cli.utils.airflow import retrieve_airflow_database_conn_from_config, set_airflow_database_conn

    project_dir_absolute = project_dir.resolve() if project_dir else Path.cwd()
    project = Project(project_dir_absolute)
    project.load_config(environment=env)

    # Since we are using the Airflow ORM to interact with connections, we need to tell Airflow to use our airflow.db
    # The usual route is to set $AIRFLOW_HOME before Airflow is imported. However, in the context of the SQL CLI, we
    # decide this during runtime, depending on the project path and SQL CLI configuration.
    airflow_meta_conn = retrieve_airflow_database_conn_from_config(project.directory / project.airflow_home)
    set_airflow_database_conn(airflow_meta_conn)

    rprint(f"Validating connection(s) for environment '{env}'")
    validate_connections(connections=project.connections, connection_id=connection)


@app.command(
    cls=AstroCommand,
    help="""
    Run a workflow locally. This task assumes that there is a local airflow DB (can be a SQLite file), that has been
    initialized with Airflow tables.
    """,
)
def run(
    workflow_name: str = typer.Argument(
        default=...,
        show_default=False,
        help="name of the workflow directory within workflows directory.",
    ),
    env: str = typer.Option(
        metavar="environment",
        default="default",
        help="environment to run in",
    ),
    project_dir: Path = typer.Option(
        None, dir_okay=True, metavar="PATH", help="(Optional) Default: current directory.", show_default=False
    ),
    generate_tasks: bool = typer.Option(
        default=True,
        help="whether to explicitly generate the tasks in your SQL CLI DAG",
        show_default=True,
    ),
    verbose: bool = typer.Option(False, help="Whether to show airflow logs", show_default=True),
) -> None:
    from airflow.utils.state import State

    from sql_cli import run_dag as dag_runner
    from sql_cli.project import Project
    from sql_cli.utils.airflow import (
        get_dag,
        retrieve_airflow_database_conn_from_config,
        set_airflow_database_conn,
    )

    project_dir_absolute = project_dir.resolve() if project_dir else Path.cwd()
    project = Project(project_dir_absolute)
    project.update_config(environment=env)
    project.load_config(env)

    # Since we are using the Airflow ORM to interact with connections, we need to tell Airflow to use our airflow.db
    # The usual route is to set $AIRFLOW_HOME before Airflow is imported. However, in the context of the SQL CLI, we
    # decide this during runtime, depending on the project path and SQL CLI configuration.
    airflow_meta_conn = retrieve_airflow_database_conn_from_config(project.directory / project.airflow_home)
    set_airflow_database_conn(airflow_meta_conn)

    dag_file = _generate_dag(project=project, workflow_name=workflow_name, generate_tasks=generate_tasks)
    dag = get_dag(dag_id=workflow_name, subdir=dag_file.parent.as_posix(), include_examples=False)

    rprint(f"\nRunning the workflow [bold blue]{dag.dag_id}[/bold blue] for [bold]{env}[/bold] environment\n")
    try:
        dr = dag_runner.run_dag(
            dag,
            run_conf=project.airflow_config,
            connections={c.conn_id: c for c in project.connections},
            verbose=verbose,
        )
    except ConnectionFailed as connection_failed:
        rprint(
            f"  [bold red]{connection_failed}[/bold red] using connection [bold]{connection_failed.conn_id}[/bold]"
        )
        raise Exit(code=1)
    except Exception as exception:
        rprint(f"  [bold red]{exception}[/bold red]")
        raise Exit(code=1)
    final_state = dr.state
    if dr.state == State.SUCCESS:
        final_state = "[bold green]SUCCESS 🚀[/bold green]"
    elif dr.state == State.FAILED:
        final_state = "[bold red]FAILED 💥[/bold red]"
    rprint(f"Completed running the workflow {dr.dag_id}. Final state: {final_state}")
    elapsed_seconds = (dr.end_date - dr.start_date).microseconds / 10**6
    rprint(f"Total elapsed time: [bold blue]{elapsed_seconds:.2}s[/bold blue]")


@app.command(
    cls=AstroCommand,
    help="""
    Initialise a project structure to write workflows using SQL files.

    Examples of usage:
    $ flow init
    $ flow init .
    $ flow init project_name

    By default, the project structure includes:
    ├── config: withholds configuration, e.g. database connections, within each environment directory
    ├── data: directory which contains datasets, including SQLite databases used by the examples
    └── workflows: directory where SQL workflows are declared, by default has two examples of workflow

    Next steps:
    * Update the file `config/default/configuration.yaml` to declare database connections.
    * Create SQL workflows within the `workflows` folder.
    """,
)
def init(
    project_dir: Path = typer.Argument(
        None, dir_okay=True, metavar="PATH", help="(Optional) Default: current directory.", show_default=False
    ),
    airflow_home: Path = typer.Option(
        None,
        dir_okay=True,
        help=f"(Optional) Set the Airflow Home. Default: {DEFAULT_AIRFLOW_HOME}",
        show_default=False,
    ),
    airflow_dags_folder: Path = typer.Option(
        None,
        dir_okay=True,
        help=f"(Optional) Set the DAGs Folder. Default: {DEFAULT_DAGS_FOLDER}",
        show_default=False,
    ),
) -> None:
    from sql_cli.project import Project

    project_dir_absolute = project_dir.resolve() if project_dir else Path.cwd()
    project = Project(project_dir_absolute, airflow_home, airflow_dags_folder)
    project.initialise()
    rprint("Initialized an Astro SQL project at", project.directory)


def _generate_dag(project: Project, workflow_name: str, generate_tasks: bool) -> Path:
    """
    Helper function for generating DAGs with proper exceptions. Moved here since this function is used by
    multiple commands

    :param project: project object containing project metadata
    :param workflow_name: name of the workflow directory for this DAG
    :param generate_tasks: whether to render each task individually or use the render function
    :return: the path to the generated dag_file
    """
    from sql_cli import dag_generator
    from sql_cli.utils.airflow import check_for_dag_import_errors

    try:
        dag_file = dag_generator.generate_dag(
            directory=project.directory / project.workflows_directory / workflow_name,
            dags_directory=project.airflow_dags_folder,
            generate_tasks=generate_tasks,
        )
    except EmptyDag:
        rprint(f"[bold red]The workflow {workflow_name} does not have any SQL files![/bold red]")
        raise Exit(code=1)
    except WorkflowFilesDirectoryNotFound:
        rprint(f"[bold red]The workflow {workflow_name} does not exist![/bold red]")
        raise Exit(code=1)
    except DagCycle as dag_cycle:
        rprint(f"[bold red]The workflow {workflow_name} contains a cycle! {dag_cycle}[/bold red]")
        raise Exit(code=1)
    import_errors = check_for_dag_import_errors(dag_file)
    if import_errors:
        all_errors = "\n\n".join(list(import_errors.values()))
        rprint(f"[bold red]Workflow failed to render[/bold red]\n errors found:\n\n {all_errors}")
        raise Exit(code=1)
    return dag_file


if __name__ == "__main__":  # pragma: no cover
    app()



import click
import typer
from typer.core import TyperCommand

from sql_cli.astro.utils import resolve_command_path


class AstroCommand(TyperCommand):
    def format_usage(self, ctx: typer.Context, formatter: click.HelpFormatter) -> None:
        """Writes the usage line into the formatter.

        This is a low-level method called by :meth:`get_usage`.
        """
        pieces = self.collect_usage_pieces(ctx)
        formatter.write_usage(resolve_command_path(ctx.command_path), " ".join(pieces))



import os


def resolve_command_path(command_path: str) -> str:
    """
    Resolve the command path using the current environment.

    :param command_path: The command path to resolve.

    :returns: the resolved command path.
    """
    if os.getenv("ASTRO_CLI"):  # being set in astro-cli dockerfile for sql-cli
        return " ".join(["astro", command_path])
    return command_path



from __future__ import annotations

import importlib
import logging
import os
from configparser import ConfigParser
from pathlib import Path
from typing import TYPE_CHECKING

from packaging.version import Version

if TYPE_CHECKING:
    from airflow.models.dag import DAG  # pragma: no cover

logger = logging.getLogger(__name__)


def airflow_version() -> Version:
    """
    Return the version of Airflow installed.
    """
    import airflow  # skipcq: PYL-W0406

    return Version(airflow.__version__)


def retrieve_airflow_database_conn_from_config(airflow_home: Path) -> str:
    """
    Retrieve the path to the Airflow metadata database connection.

    :params airfow_home: Path to where Airflow was initialized ($AIRFLOW_HOME).
    :returns: Airflow metadata database connection URI
    """
    os.environ["AIRFLOW_HOME"] = str(airflow_home.resolve())
    filename = airflow_home / "airflow.cfg"
    parser = ConfigParser()
    parser.read(filename)
    confdict = {section: dict(parser.items(section)) for section in parser.sections()}

    if airflow_version() >= Version("2.3"):
        sql_alchemy_conn = confdict["database"]["sql_alchemy_conn"]
    else:
        sql_alchemy_conn = confdict["core"]["sql_alchemy_conn"]

    return sql_alchemy_conn


def disable_examples(airflow_home: Path) -> None:
    """
    Disable Airflow examples in the configuration file available at the given airflow_home directory.

    :params airflow_home: Path to where Airflow was initialised ($AIRFLOW_HOME)
    """
    filename = airflow_home / "airflow.cfg"
    parser = ConfigParser()
    parser.read(filename)
    parser["core"]["load_examples"] = "False"
    parser["core"]["logging_level"] = "WARN"
    with open(filename, "w") as out_fp:
        parser.write(out_fp)


def set_airflow_database_conn(airflow_meta_conn: str) -> None:
    """
    Given a desired Airflow DB connection string, refresh Airflow settings so
    that Airflow ORM uses this database as metadata store.

    :params airflow_db_conn: Similar to `sqlite:////tmp/project/airflow.db`
    """
    if airflow_version() >= Version("2.3"):
        # This is a hacky approach we managed to find to make things work with Airflow 2.4
        os.environ["AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"] = airflow_meta_conn
    else:
        os.environ["AIRFLOW__CORE__SQL_ALCHEMY_CONN"] = airflow_meta_conn

    import airflow  # skipcq: PYL-W0406

    importlib.reload(airflow)
    importlib.reload(airflow.configuration)
    importlib.reload(airflow.models.base)
    importlib.reload(airflow.models.connection)


# The following function was copied from Apache Airflow
# https://github.com/apache/airflow/commit/ce071172e22fba018889db7dcfac4a4d0fc41cda
# And we should replace by the upstream method once Airflow 2.5 is released
# We are copying it so that we do not include examples
# This helps silencing the SQL CLI output and also the speed of the run command
def _search_for_dag_file(val: str) -> str:
    """
    Search for the file referenced at fileloc.
    By the time we get to this function, we've already run this `val` through `process_subdir`
    and loaded the DagBag there and came up empty.  So here, if `val` is a file path, we make
    a last ditch effort to try and find a dag file with the same name in our dags folder. (This
    avoids the unnecessary dag parsing that would occur if we just parsed the dags folder).
    If `val` is a path to a file, this likely means that the serializing process had a dags_folder
    equal to only the dag file in question. This prevents us from determining the relative location.
    And if the paths are different between worker and dag processor / scheduler, then we won't find
    the dag at the given location.
    """
    from airflow import settings

    if val and Path(val).suffix in (".zip", ".py"):
        matches = list(Path(settings.DAGS_FOLDER).rglob(Path(val).name))
        if len(matches) == 1:
            return matches[0].as_posix()
    return ""


# The following function was copied from Apache Airflow
# https://github.com/apache/airflow/commit/ce071172e22fba018889db7dcfac4a4d0fc41cda
# And we should replace by the upstream method once Airflow 2.5 is released
# We are copying it so that we do not include examples
# This helps silencing the SQL CLI output and also the speed of the run command
def get_dag(subdir: str, dag_id: str, include_examples: bool = False) -> DAG:
    """
    Returns DAG of a given dag_id
    First it we'll try to use the given subdir.  If that doesn't work, we'll try to
    find the correct path (assuming it's a file) and failing that, use the configured
    dags folder.
    """
    from airflow import settings
    from airflow.exceptions import AirflowException
    from airflow.models import DagBag
    from airflow.utils.cli import process_subdir

    first_path = process_subdir(subdir)
    dagbag = DagBag(first_path, include_examples=include_examples)
    if dag_id not in dagbag.dags:
        fallback_path = _search_for_dag_file(subdir) or settings.DAGS_FOLDER
        logger.warning("Dag %r not found in path %s; trying path %s", dag_id, first_path, fallback_path)
        dagbag = DagBag(dag_folder=fallback_path, include_examples=include_examples)
        if dag_id not in dagbag.dags:
            raise AirflowException(
                f"Dag {dag_id!r} could not be found; either it does not exist or it failed to parse."
            )
    return dagbag.dags[dag_id]


def check_for_dag_import_errors(dag_file: Path) -> dict[str, str]:
    """
    Check for import errors in DAG

    :param dag_file: The path to the dag file.

    :returns: the import errors found per DAG during processing the DagBag.
    """
    from airflow.models import DagBag
    from airflow.utils.cli import process_subdir

    dag_folder = process_subdir(str(dag_file))
    dagbag = DagBag(dag_folder, include_examples=False)
    return dagbag.import_errors



from __future__ import annotations

from pathlib import Path

from airflow.models import Connection
from airflow.utils.session import create_session

from sql_cli.utils.rich import rprint

CONNECTION_ID_OUTPUT_STRING_WIDTH = 25


def _create_or_replace_connection(conn_obj: Connection) -> None:
    """Creates a new or replaces existing connection in the Airflow DB with the given connection object."""
    conn_id = conn_obj.conn_id
    with create_session() as session:
        db_connection = session.query(Connection).filter_by(conn_id=conn_id).one_or_none()
        if db_connection:
            session.delete(db_connection)
            session.commit()
        session.add(conn_obj)
        session.commit()


def validate_connections(connections: list[Connection], connection_id: str | None = None) -> None:
    """
    Validates that the given connections are valid and registers them to Airflow with replace policy for existing
    connections.
    """
    config_file_contains_connection = False

    for connection in connections:
        if connection.id == connection_id:
            config_file_contains_connection = True
        _create_or_replace_connection(connection)
        status = "[bold green]PASSED[/bold green]" if _is_valid(connection) else "[bold red]FAILED[/bold red]"
        rprint(f"Validating connection {connection.conn_id:{CONNECTION_ID_OUTPUT_STRING_WIDTH}}", status)

    if not config_file_contains_connection:
        rprint("Error: Config file does not contain given connection", connection_id)


def _is_valid(connection: Connection) -> bool:
    # Sqlite automatically creates the file if it does not exist,
    # but our users might not expect that. They are referencing a database they expect to exist.
    if connection.conn_type == "sqlite" and not Path(connection.host).is_file():
        return False

    success_status, _ = connection.test_connection()
    return success_status



from __future__ import annotations

from typing import Any

from astro.custom_backend.serializer import deserialize
from astro.sql import LoadFileOperator


def build_load_file_operator_content(
    operator_content: dict[str, Any],
    operator_file_name: str,
) -> dict[str, Any]:
    """
    Deserialize the File and Table classes.

    :param operator_content: The dictionary of the operator content to be deserialized.
    :param operator_file_name: The file name of the operator.

    :returns: the operator content deserialized.
    """
    input_file_content = operator_content.pop("input_file")
    input_file_content["class"] = "File"
    input_file_content.setdefault("conn_id", None)
    input_file_content.setdefault("is_dataframe", False)
    input_file_content.setdefault("normalize_config", None)

    operator_content["input_file"] = deserialize(input_file_content)

    output_table_content = operator_content.pop("output_table")
    output_table_content["class"] = "Table"
    output_table_content.setdefault("metadata", {})
    output_table_content.setdefault("name", operator_file_name)
    output_table_content.setdefault("temp", False)

    operator_content["output_table"] = deserialize(output_table_content)

    return operator_content


def get_load_file_instance(yaml_content: dict[str, Any], operator_file_name: str) -> LoadFileOperator:
    """
    Instantiate the LoadFileOperator.

    :param yaml_content: The content of the load_file yaml.
    :param operator_file_name: The file name of the operator.

    :returns: an instance of the LoadFileOperator
    """
    operator_content: dict[str, Any] = build_load_file_operator_content(yaml_content, operator_file_name)
    return LoadFileOperator(**operator_content)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from networkx import DiGraph, depth_first_search, find_cycle, is_directed_acyclic_graph

from sql_cli.exceptions import DagCycle, EmptyDag, WorkflowFilesDirectoryNotFound
from sql_cli.utils.jinja import render
from sql_cli.workflow_directory_parser import WorkflowFile, get_workflow_files

TEMPLATES_DIRECTORY = Path("templates")


@dataclass(frozen=True)
class Workflow:
    """
    A DAG of workflow files i.e. used for finding the right order to execute the files in.

    :param dag_id: The id of the DAG to generate.
    :param start_date: The start date of the DAG.
    :param workflow_files: The workflow files to use for DAG generation.
    """

    dag_id: str
    start_date: datetime
    workflow_files: list[WorkflowFile]

    def __post_init__(self) -> None:
        if not self.workflow_files:
            raise EmptyDag("Missing workflow files!")

    def has_workflow_file(self, variable_name: str) -> bool:
        """
        Check whether the given variable name belongs to a real workflow file.

        :params variable_name: The variable name of the workflow file to check.

        :returns: True if there is any workflow file with the given variable name.
        """
        return any(
            workflow_file.get_variable_name() == variable_name for workflow_file in self.workflow_files
        )

    def find_workflow_file(self, variable_name: str) -> WorkflowFile:
        """
        Find a SQL file with the given variable name.

        :params variable_name: The variable name of the workflow file to find.

        :returns: if found a workflow file else raises an exception.
        """
        try:
            return next(
                workflow_file
                for workflow_file in self.workflow_files
                if workflow_file.get_variable_name() == variable_name
            )
        except StopIteration:
            raise ValueError("No workflow file has been found for variable name!")

    def sorted_workflow_files(self) -> list[WorkflowFile]:
        """
        Build, validate and sort the SQL files.

        :returns: a list of sql files sorted
            so they can be called sequentially in python code.
        """
        # Create a graph based on all sql files and parameters
        graph = DiGraph(
            [
                (workflow_file, self.find_workflow_file(parameter))
                for workflow_file in self.workflow_files
                for parameter in workflow_file.get_parameters()
                if self.has_workflow_file(parameter)
            ]
        )

        if not graph.nodes:
            # Add nodes without edges i.e. without any table references
            graph.add_nodes_from(self.workflow_files)

        if not is_directed_acyclic_graph(graph):
            cycle_edges = " and ".join(
                " and ".join(edge.get_variable_name() for edge in edges) for edges in find_cycle(graph)
            )
            raise DagCycle(f"A cycle between {cycle_edges} has been detected!")

        return list(depth_first_search.dfs_postorder_nodes(graph))


def generate_dag(directory: Path, dags_directory: Path, generate_tasks: bool) -> Path:
    """
    Generate a DAG from SQL files.

    :params directory: The directory containing the raw workflow files.
    :params dags_directory: The directory containing the generated DAG.
    :params generate_tasks: Whether the user wants to explicitly generate each task
        of the airflow DAG or rely on a less verbose `render` function.

    :returns: the path to the DAG file.
    """
    if not directory.exists():
        raise WorkflowFilesDirectoryNotFound("The directory does not exist!")
    workflow_files = sorted(get_workflow_files(directory, target_directory=dags_directory))
    workflow_files_dag = Workflow(
        dag_id=directory.name,
        start_date=datetime(2020, 1, 1),
        workflow_files=workflow_files,
    )
    if generate_tasks:
        template_file = TEMPLATES_DIRECTORY / "gen_tasks_dag.py.jinja2"
    else:
        for workflow_file in workflow_files_dag.sorted_workflow_files():
            workflow_file.write_raw_content_to_target_path()
        template_file = TEMPLATES_DIRECTORY / "render_dag.py.jinja2"

    output_file = dags_directory / f"{workflow_files_dag.dag_id}.py"
    render(
        template_file=template_file,
        context={"dag": workflow_files_dag},
        output_file=output_file,
    )
    return output_file



from __future__ import annotations

import os
import shutil
from configparser import ConfigParser
from pathlib import Path
from typing import Any

from airflow.models.connection import Connection
from packaging.version import Version

from sql_cli.configuration import Config, convert_to_connection
from sql_cli.constants import DEFAULT_AIRFLOW_HOME, DEFAULT_DAGS_FOLDER, DEFAULT_ENVIRONMENT
from sql_cli.exceptions import InvalidProject
from sql_cli.utils.airflow import airflow_version, disable_examples

BASE_SOURCE_DIR = Path(os.path.realpath(__file__)).parent.parent / "include/base/"

MANDATORY_PATHS = {
    Path("config/default/configuration.yml"),
    Path("config/global/configuration.yml"),
    Path("workflows"),
    Path(".airflow/default/airflow.db"),
}


class Project:
    """
    SQL CLI Project.
    """

    workflows_directory = Path("workflows")

    def __init__(
        self,
        directory: Path,
        airflow_home: Path | None = None,
        airflow_dags_folder: Path | None = None,
    ) -> None:
        self.directory = directory
        self._airflow_home = airflow_home or Path(self.directory, DEFAULT_AIRFLOW_HOME)
        self._airflow_dags_folder = airflow_dags_folder or Path(self.directory, DEFAULT_DAGS_FOLDER)
        self.connections: list[Connection] = []

    @property
    def airflow_home(self) -> Path:
        """
        Folder which contains the Airflow database and configuration.
        Can be either user-defined, during initialisation, or the default one.

        This is used by flow validate and flow run.

        :returns: The path to the Airflow home directory.
        """
        return self._airflow_home

    @property
    def airflow_dags_folder(self) -> Path:
        """
        Folder which contains the generated Airflow DAG files.
        Can be eitehr user-defined, during initialisation, or the default one.

        This is used by flow generate and flow run.

        :returns: The path to the Airflow DAGs directory.
        """
        return self._airflow_dags_folder

    @property
    def airflow_config(self) -> dict[str, Any]:
        """
        Retrieve the Airflow configuration for the currently set environment.

        :returns: A Python dictionary containing the Airflow configuration.
        """
        filename = self.airflow_home / "airflow.cfg"
        parser = ConfigParser()
        parser.read(filename)
        return {section: dict(parser.items(section)) for section in parser.sections()}

    def update_config(self, environment: str = DEFAULT_ENVIRONMENT) -> None:
        """
        Sets custom Airflow configuration in case the user is not using the default values.

        :param environment: the environment for which the configuration has to be updated
        """
        config = Config(environment=environment, project_dir=self.directory)
        config = config.from_yaml_to_config()

        global_env_filepath = config.get_global_config_filepath()
        if self._airflow_home is not None:
            config.write_value_to_yaml(
                "airflow", "home", str(self._airflow_home.resolve()), global_env_filepath
            )
        if self._airflow_dags_folder is not None:
            config.write_value_to_yaml(
                "airflow", "dags_folder", str(self._airflow_dags_folder.resolve()), global_env_filepath
            )

        config.connections[0]["host"] = str(self.directory / config.connections[0]["host"])
        config.write_config_to_yaml()

    def _initialise_airflow(self) -> None:
        """
        Create an Airflow database and configuration in the self.airflow_home folder, or upgrade them,
        if they already exist.
        """
        os.environ.pop("AIRFLOW_HOME", None)
        if airflow_version() >= Version("2.3"):
            os.environ.pop("AIRFLOW__DATABASE__SQL_ALCHEMY_CONN", None)
        else:
            os.environ.pop("AIRFLOW__CORE__SQL_ALCHEMY_CONN", None)

        # TODO: In future we want to replace this by either:
        # - python-native approach or
        # - subprocess
        os.system(  # skipcq: BAN-B605
            f"AIRFLOW_HOME={self.airflow_home} "
            f"AIRFLOW__CORE__DAGS_FOLDER={self.airflow_dags_folder} "
            "AIRFLOW__CORE__LOAD_EXAMPLES=False "
            "airflow db init > /dev/null 2>&1"
        )

    def _remove_unnecessary_airflow_files(self) -> None:
        """
        Delete Airflow generated paths which are not necessary for the SQL CLI (scheduler & webserver-related).
        """
        logs_folder = self.airflow_home / "logs"
        shutil.rmtree(logs_folder)

        webserver_config = self.airflow_home / "webserver_config.py"
        webserver_config.unlink()

    def initialise(self) -> None:
        """
        Initialise a SQL CLI project, creating expected directories and files.

        :param airflow_home: Custom user-defined Airflow Home directory
        :param airflow_dags_folder: Custom user-defined Airflow DAGs folder
        """
        shutil.copytree(
            src=BASE_SOURCE_DIR,
            dst=self.directory,
            ignore=shutil.ignore_patterns(".gitkeep"),
            dirs_exist_ok=True,
        )
        self.update_config()
        self._initialise_airflow()
        disable_examples(self.airflow_home)
        self._remove_unnecessary_airflow_files()

    def is_valid_project(self) -> bool:
        """
        Check if self.directory contains the necessary paths which make it qualify as a valid SQL CLI project.

        The mandatory paths are sql_cli.project.MANDATORY_PATHS
        """
        existing_paths = {path.relative_to(self.directory) for path in Path(self.directory).rglob("*")}
        return MANDATORY_PATHS.issubset(existing_paths)

    def load_config(self, environment: str = DEFAULT_ENVIRONMENT) -> None:
        """
        Given a self.directory and an environment, load to the configuration ad paths to the Project instance.

        :param environment: string referencing the desired environment, uses "default" unless specified
        """
        if not self.is_valid_project():
            raise InvalidProject("This is not a valid SQL project. Please, use `flow init`")
        config = Config(environment=environment, project_dir=self.directory).from_yaml_to_config()
        if config.airflow_home:
            self._airflow_home = Path(config.airflow_home).resolve()
        if config.airflow_dags_folder:
            self._airflow_dags_folder = Path(config.airflow_dags_folder).resolve()
        self.connections = [convert_to_connection(c) for c in config.connections]



from __future__ import annotations

from abc import abstractmethod
from pathlib import Path
from typing import Any, Iterable

import airflow
import frontmatter
import yaml
from airflow.models import BaseOperator
from yaml.scanner import ScannerError

from astro.sql import LoadFileOperator
from astro.sql.operators.transform import TransformOperator
from sql_cli.constants import GENERATED_WORKFLOW_INCLUDE_DIRECTORY
from sql_cli.operators.load_file import get_load_file_instance
from sql_cli.utils.jinja import find_template_variables


class WorkflowFile:
    """
    A Workflow is equivalent to a chain of corresponding operators in Astro SDK.

    :param root_directory: The root directory path of the project.
    :param path: The path to the workflow file.
    :param target_directory: The target directory path for the executable workflow file.
    """

    def __init__(self, root_directory: Path, path: Path, target_directory: Path) -> None:
        self.root_directory = root_directory
        self.path = path
        self.target_directory = target_directory
        self.raw_content = self.path.read_text()
        self.metadata, self.content = frontmatter.parse(self.raw_content)

    def __eq__(self, other: object) -> bool:
        """
        Check if this workflow file equals the given object.

        :param other: The object to compare to.

        :returns: True if this workflow file is equal to the other one.
        """
        if isinstance(other, WorkflowFile):
            return self.root_directory == other.root_directory and self.path == other.path
        return False

    def __gt__(self, other: WorkflowFile) -> bool:
        """
        Check if this workflow file path comes before the other workflow file path in alphabetic order.

        :param other: The other workflow file to compare to.

        :returns: True if this workflow file path comes before the other one in alphabetic order.
        """
        return self.root_directory / self.path > self.root_directory / other.path

    def __hash__(self) -> int:
        """
        Defines the hash for the workflow file.

        :returns: the hash of the workflow file.
        """
        return hash(self.root_directory) ^ hash(self.path)

    @property
    def name(self) -> str:
        """
        The file name of the workflow file.

        :returns the file name without last suffix.
        """
        return self.path.stem

    @property
    def operator_name(self) -> str:
        """
        The name of the aql function to run.

        :returns: the function name of the operator in aql.
        """
        raise NotImplementedError("Operator name is not defined.")

    def get_parameters(self) -> list[str]:
        """
        Get all parameters used for parameterized workflow files.

        :returns: declared parameters for the workflow file.
        """
        return sorted(find_template_variables(self.path))

    def has_sub_directory(self) -> bool:
        """
        Check if workflow file is in a subdirectory.

        :returns: True if it's not in the parent root directory.
        """
        return self.path.parent != self.root_directory

    def get_sub_directories(self) -> Iterable[str]:
        """
        Get the directory names between root and workflow file path.

        :yields: a sub directory name.
        """
        for parent in self.path.parents:
            if parent == self.root_directory:
                break
            yield parent.name

    def get_variable_name(self) -> str:
        """
        Get the variable name used as a unique identifier as a python variable.

        :returns: the file name without suffix.
        """
        if self.has_sub_directory():
            return f"{'__'.join(self.get_sub_directories())}__{self.name}"
        return self.name

    def get_relative_target_path(self) -> Path:
        """
        Get the relative path to the executable workflow file within the DAGs folder.

        :returns: the path where workflow files without any headers are being placed.
        """
        target_full_directory = (
            self.target_directory
            / GENERATED_WORKFLOW_INCLUDE_DIRECTORY
            / self.root_directory.name
            / "/".join(self.get_sub_directories())
        )
        target_full_directory.mkdir(parents=True, exist_ok=True)

        target_path = target_full_directory / self.path.name

        target_path.write_text(self.content)

        return target_path.relative_to(self.target_directory)

    def write_raw_content_to_target_path(self) -> None:
        """
        Writes both content and headers to the target directory.
        This is because with the "render" function, we will still need
        the headers for creating proper TransformOperators
        """
        target_full_directory = (
            self.target_directory
            / GENERATED_WORKFLOW_INCLUDE_DIRECTORY
            / self.root_directory.name
            / "/".join(self.get_sub_directories())
        )
        target_full_directory.mkdir(parents=True, exist_ok=True)

        target_path = target_full_directory / self.path.name

        target_path.write_text(self.raw_content)

    @abstractmethod
    def to_operator(self) -> BaseOperator:
        """
        Deserialize the operator to an actual aql operator.

        :returns: the aql operator.
        """
        raise NotImplementedError("aql operator not implemented!")


class SqlFile(WorkflowFile):
    """A SqlFile is equivalent to a transform step in the Astro SDK."""

    @property
    def operator_name(self) -> str:
        """
        The name of the aql function to run.

        :returns: the function name of the operator in aql.
        """
        return "transform_file"

    def to_operator(self) -> TransformOperator:
        """
        Converts SQLFile into a TransformOperator that can be added to a DAG.
        Any relevant metadata from the file frontmatter will be passed to the TransformOperator,
        though we do not pass parameter dependencies at this stage.

        :return: a TransformOperator
        """
        kwargs = {
            "conn_id": self.metadata.get("conn_id"),
            "parameters": None,
            "handler": None,
            "database": self.metadata.get("database"),
            "schema": self.metadata.get("schema"),
            "python_callable": lambda: (str(self.path), None),
            "sql": self.content,
        }
        if airflow.__version__.startswith("2.2."):
            kwargs["op_args"] = []
        return TransformOperator(**kwargs)


class YamlFile(WorkflowFile):
    """
    A YamlFile, for example, could be a load_file step in the Astro SDK.

    :param root_directory: The root directory path of the project.
    :param path: The path to the yaml file.
    :param target_directory: The target directory path for the executable yaml file.
    """

    SUPPORTED_OPERATORS = {"load_file": get_load_file_instance}

    def __init__(self, root_directory: Path, path: Path, target_directory: Path) -> None:
        super().__init__(root_directory, path, target_directory)
        self.yaml_content = yaml.safe_load(self.raw_content)

    @property
    def operator_name(self) -> str:
        """
        The name of the aql function to run.

        :returns: the function name of the operator in aql.
        """
        operators = list(self.yaml_content.keys())
        operators_count = len(operators)

        if operators_count != 1:
            raise ScannerError(f"Exactly one operator expected. Got {operators_count} operators: {operators}")

        operator = operators[0]

        if operator not in self.SUPPORTED_OPERATORS:
            raise NotImplementedError(f"Operator support for {operator} not available!")

        return operator

    def get_yaml_content(self) -> dict[str, Any]:
        """
        Get the yaml content for the operator.

        :returns: the operator kwargs.
        """
        return self.yaml_content[self.operator_name]

    def to_operator(self) -> LoadFileOperator:
        """
        Deserialize the operator to an actual aql operator.

        :returns: the aql operator.
        """
        return self.SUPPORTED_OPERATORS[self.operator_name](self.get_yaml_content(), self.name)


SUPPORTED_FILES = {"*.sql": SqlFile, "*.yaml": YamlFile, "*.yml": YamlFile}


def get_workflow_files(directory: Path, target_directory: Path | None) -> set[WorkflowFile]:
    """
    Get all workflow files within a directory.

    :param directory: The directory look in for files.
    :param target_directory: The target directory path for the executable workflow.

    :returns: a set of all the workflow files.
    """
    return {
        SUPPORTED_FILES[file_type](
            root_directory=directory,
            path=child,
            target_directory=target_directory,  # type: ignore
        )
        for file_type in SUPPORTED_FILES
        for child in directory.rglob(file_type)
        if child.is_file() and not child.is_symlink()
    }



CONFIG_DIR = "config"
GLOBAL_CONFIG_DIRECTORY = "global"
CONFIG_FILENAME = "configuration.yml"

DEFAULT_ENVIRONMENT = "default"
DEFAULT_AIRFLOW_HOME = f".airflow/{DEFAULT_ENVIRONMENT}"
DEFAULT_DAGS_FOLDER = ".airflow/dags"

GENERATED_WORKFLOW_INCLUDE_DIRECTORY = "include"



from typing import IO, Any, Optional, Union

import click
from rich.highlighter import ReprHighlighter
from rich.panel import Panel
from typer.rich_utils import (
    ALIGN_ERRORS_PANEL,
    ERRORS_PANEL_TITLE,
    STYLE_ERRORS_PANEL_BORDER,
    STYLE_ERRORS_SUGGESTION,
    _get_rich_console,
    highlighter,
)

from sql_cli.astro.utils import resolve_command_path


def rich_format_error(self: click.ClickException) -> None:
    """Print richly formatted click errors.

    Called by custom exception handler to print richly formatted click errors.
    Mimics original click.ClickException.echo() function but with rich formatting.
    """
    console = _get_rich_console(stderr=True)
    ctx: Union[click.Context, None] = getattr(self, "ctx", None)
    if ctx is not None:
        console.print(ctx.get_usage())

    if ctx is not None and ctx.command.get_help_option(ctx) is not None:
        console.print(
            f"Try [blue]'{resolve_command_path(ctx.command_path)} {ctx.help_option_names[0]}'[/] for help.",
            style=STYLE_ERRORS_SUGGESTION,
        )

    console.print(
        Panel(
            highlighter(self.format_message()),
            border_style=STYLE_ERRORS_PANEL_BORDER,
            title=ERRORS_PANEL_TITLE,
            title_align=ALIGN_ERRORS_PANEL,
        )
    )


def rprint(
    *objects: Any,
    sep: str = " ",
    end: str = "\n",
    file: Optional[IO[str]] = None,
    flush: bool = False,  # skipcq: PYL-W0613
) -> None:
    r"""Print object(s) supplied via positional arguments.
    This function has an identical signature to the built-in print.
    For more advanced features, see the :class:`~rich.console.Console` class.

    Args:
        sep (str, optional): Separator between printed objects. Defaults to " ".
        end (str, optional): Character to write at end of output. Defaults to "\\n".
        file (IO[str], optional): File to write to, or None for stdout. Defaults to None.
        flush (bool, optional): Has no effect as Rich always flushes output. Defaults to False.

    """
    from rich.console import Console

    # We use the rich console from typer which allows us to set FORCE_TERMINAL & MAX_WIDTH
    # both we use for consistent output locally (via Makefile) and in CI
    console = _get_rich_console()
    # And we reset the hightlighter to the default highlighter,
    # because the OptionHighlighter only makes sense for options e.g in help output
    console.highlighter = ReprHighlighter()

    write_console = console if file is None else Console(file=file)
    return write_console.print(*objects, sep=sep, end=end)



from __future__ import annotations

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import yaml
from airflow.models.connection import Connection

from sql_cli.constants import CONFIG_DIR, CONFIG_FILENAME, GLOBAL_CONFIG_DIRECTORY


def convert_to_connection(conn: dict[str, Any]) -> Connection:
    """
    Convert the SQL CLI connection dictionary into an Airflow Connection instance.

    :param conn: SQL CLI connection dictionary
    :returns: Connection object
    """
    from airflow.api_connexion.schemas.connection_schema import connection_schema

    c = conn.copy()
    c["connection_id"] = c["conn_id"]
    c.pop("conn_id")
    return Connection(**connection_schema.load(c))


@dataclass
class Config:
    """
    SQL CLI Project configuration class to support reading and writing to configuration file.
    """

    project_dir: Path
    environment: str

    connections: list[dict[str, Connection]] = field(default_factory=list)
    airflow_home: str | None = None
    airflow_dags_folder: str | None = None

    def get_env_config_filepath(self) -> Path:
        """
        Return environment specific configuration.yaml filepath.

        :returns: The path to the desired YAML configuration file
        """
        return self.project_dir / CONFIG_DIR / self.environment / CONFIG_FILENAME

    def get_global_config_filepath(self) -> Path:
        """
        Return global configuration.yaml filepath which is shared across environments.

        :return: The path to the desired global YAML configuration file
        """
        return self.project_dir / CONFIG_DIR / GLOBAL_CONFIG_DIRECTORY / CONFIG_FILENAME

    @staticmethod
    def from_yaml_to_dict(filepath: Path) -> dict[str, Any]:
        """
        Return a dict with the contents of the given configuration.yaml

        :param filepath: Path of the desired configuration.yaml to read contents from.

        :returns: Content of the YAML configuration file as a python dictionary.
        """
        with open(filepath) as fp:
            yaml_with_env = os.path.expandvars(fp.read())
            yaml_config = yaml.safe_load(yaml_with_env)
        return yaml_config

    def from_yaml_to_config(self) -> Config:
        """Returns a Config instance with the contents of the environment specific and global configuration.yaml"""
        env_yaml_config = self.from_yaml_to_dict(self.get_env_config_filepath())
        global_yaml_config = self.from_yaml_to_dict(self.get_global_config_filepath()) or {}
        return Config(
            project_dir=self.project_dir,
            environment=self.environment,
            airflow_home=global_yaml_config.get("airflow", {}).get("home"),
            airflow_dags_folder=global_yaml_config.get("airflow", {}).get("dags_folder"),
            connections=env_yaml_config["connections"],
        )

    def write_value_to_yaml(self, section: str, key: str, value: str, filepath: Path) -> None:
        """
        Write a particular key/value to the desired configuration.yaml.

        Example:
        ```
            [section]
            key = value
        ```

        :param section: Section within the YAML file where the key/value will be recorded
        :param key: Key within the YAML file associated to the value to be recorded.
        :param value: Value associated to the key in the YAML file.
        """
        yaml_config = self.from_yaml_to_dict(filepath)
        yaml_config.setdefault(section, {})
        yaml_config[section][key] = value

        with filepath.open(mode="w") as fp:
            yaml.dump(yaml_config, fp)

    def write_config_to_yaml(self) -> None:
        """Write Config instance's key-values to respective environment specific and global configuration.yml."""
        env_config_filepath = self.get_env_config_filepath()
        env_yaml_config = self.from_yaml_to_dict(env_config_filepath)
        env_yaml_config["connections"] = self.connections
        with open(env_config_filepath, "w") as fp:
            yaml.dump(env_yaml_config, fp)

        global_config_filepath = self.get_global_config_filepath()
        global_yaml_config = self.from_yaml_to_dict(global_config_filepath)
        if self.airflow_home:
            global_yaml_config["airflow"]["home"] = self.airflow_home
        if self.airflow_dags_folder:
            global_yaml_config["airflow"]["dags_folder"] = self.airflow_dags_folder
        with open(global_config_filepath, "w") as fp:
            yaml.dump(global_yaml_config, fp)



from __future__ import annotations

import logging
import sys
import warnings
from datetime import datetime
from typing import Any, cast

from airflow import settings as airflow_settings
from airflow.models.connection import Connection
from airflow.models.dag import DAG
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.secrets.local_filesystem import LocalFilesystemBackend
from airflow.utils import timezone
from airflow.utils.session import provide_session
from airflow.utils.state import DagRunState, State
from airflow.utils.types import DagRunType
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.session import Session

from astro.sql.operators.cleanup import AstroCleanupException
from sql_cli.exceptions import ConnectionFailed
from sql_cli.utils.rich import rprint

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

# This was added to airflow.utils.session as part of Airflow 2.2.4, we're copying it here for backwwards
# compatibility (currently we support Airflow 2.1 onwards)
# It seems like a hack to avoid MyPy complaining we're assigning None to a variable that should be an Airflow session
NEW_SESSION = airflow_settings.SASession = cast(airflow_settings.SASession, None)


class AstroFilesystemBackend(LocalFilesystemBackend):
    def __init__(
        self,
        connections: dict[str, Connection] | None = None,
        variables_file_path: str | None = None,
        connections_file_path: str | None = None,
    ):
        self._local_conns: dict[str, Connection] = connections or {}
        super().__init__(variables_file_path=variables_file_path, connections_file_path=connections_file_path)

    @property
    def _local_connections(self) -> dict[str, Connection]:
        conns = self._local_conns
        conns.update(super()._local_connections)
        return conns


@provide_session
def run_dag(
    dag: DAG,
    execution_date: datetime | None = None,
    run_conf: dict[str, Any] | None = None,
    conn_file_path: str | None = None,
    variable_file_path: str | None = None,
    connections: dict[str, Connection] | None = None,
    verbose: bool = False,
    session: Session = NEW_SESSION,
) -> DagRun:
    """
    Execute one single DagRun for a given DAG and execution date.

    :param dag: The Airflow DAG we will run
    :param execution_date: execution date for the DAG run
    :param run_conf: configuration to pass to newly created dagrun
    :param conn_file_path: file path to a connection file in either yaml or json
    :param variable_file_path: file path to a variable file in either yaml or json
    :param session: database connection (optional)

    :return: the dag run object
    """
    execution_date = execution_date or timezone.utcnow()
    dag.log.debug("Clearing existing task instances for execution date %s", execution_date)
    dag.clear(
        start_date=execution_date,
        end_date=execution_date,
        dag_run_state=False,  # type: ignore
        session=session,
    )
    dag.log.debug("Getting dagrun for dag %s", dag.dag_id)
    dr: DagRun = _get_or_create_dagrun(
        dag=dag,
        start_date=execution_date,
        execution_date=execution_date,
        run_id=DagRun.generate_run_id(DagRunType.MANUAL, execution_date),
        session=session,
        conf=run_conf,
    )

    if conn_file_path or variable_file_path or connections:
        # To get around the fact that we reload certain modules, we need to import here
        from airflow.configuration import secrets_backend_list

        local_secrets = AstroFilesystemBackend(
            variables_file_path=variable_file_path,
            connections_file_path=conn_file_path,
            connections=connections,
        )
        secrets_backend_list.insert(0, local_secrets)

    tasks = dag.task_dict
    dag.log.debug("starting dagrun")
    # Instead of starting a scheduler, we run the minimal loop possible to check
    # for task readiness and dependency management. This is notably faster
    # than creating a BackfillJob and allows us to surface logs to the user
    while dr.state == State.RUNNING:
        schedulable_tis, _ = dr.update_state(session=session)
        for ti in schedulable_tis:
            if verbose:
                add_logger_if_needed(dag, ti)
            ti.task = tasks[ti.task_id]
            _run_task(ti, session=session)
    if conn_file_path or variable_file_path or connections:
        # Remove the local variables we have added to the secrets_backend_list
        secrets_backend_list.pop(0)  # noqa
    return dr


def add_logger_if_needed(dag: DAG, ti: TaskInstance) -> None:
    """
    Add a formatted logger to the taskinstance so all logs are surfaced to the command line instead
    of into a task file. Since this is a local test run, it is much better for the user to see logs
    in the command line, rather than needing to search for a log file.

    :param ti: The taskinstance that will receive a logger
    """
    logging_format = logging.Formatter("[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s")
    handler = logging.StreamHandler(sys.stdout)
    handler.level = logging.DEBUG
    handler.setFormatter(logging_format)
    # only add log handler once
    if not any(isinstance(h, logging.StreamHandler) for h in ti.log.handlers):
        dag.log.debug("Adding Streamhandler to taskinstance %s", ti.task_id)
        ti.log.addHandler(handler)


def _run_task(ti: TaskInstance, session: Session) -> None:
    """
    Run a single task instance, and push result to Xcom for downstream tasks. Bypasses a lot of
    extra steps used in `task.run` to keep our local running as fast as possible
    This function is only meant for the `dag.test` function as a helper function.

    :param ti: TaskInstance to run
    """
    if hasattr(ti, "map_index") and ti.map_index >= 0:
        rprint(f"Processing [bold yellow]{ti.task_id}[/bold yellow][{ti.map_index}]...", end=" ")
    else:
        rprint(f"Processing [bold yellow]{ti.task_id}[/bold yellow]...", end=" ")

    try:
        warnings.filterwarnings(action="ignore")
        ti._run_raw_task(session=session)  # skipcq: PYL-W0212
        session.flush()
        session.commit()
        rprint("[bold green]SUCCESS[/bold green]")
    except OperationalError as operational_exception:
        rprint("[bold red]FAILED[/bold red]")
        orig_exception = operational_exception.orig
        orig_message = orig_exception.args[0]
        raise ConnectionFailed(orig_message, conn_id=ti.task.conn_id) from orig_exception
    except AstroCleanupException:
        rprint("aql.cleanup async, continuing")
    except Exception:
        rprint("[bold red]FAILED[/bold red]")
        raise


def _get_or_create_dagrun(
    dag: DAG,
    conf: dict[Any, Any] | None,
    start_date: datetime,
    execution_date: datetime,
    run_id: str,
    session: Session,
) -> DagRun:
    """
    Create a DAGRun, but only after clearing the previous instance of said dagrun to prevent collisions.
    This function is only meant for the `dag.test` function as a helper function.

    :param dag: Dag to be used to find dagrun
    :param conf: configuration to pass to newly created dagrun
    :param start_date: start date of new dagrun, defaults to execution_date
    :param execution_date: execution_date for finding the dagrun
    :param run_id: run_id to pass to new dagrun
    :param session: sqlalchemy session

    :return: the Dagrun object needed to run tasks.
    """
    log.debug("dagrun id: %s", dag.dag_id)
    dr: DagRun = (
        session.query(DagRun)
        .filter(DagRun.dag_id == dag.dag_id, DagRun.execution_date == execution_date)
        .first()
    )
    if dr:
        session.delete(dr)
        session.commit()
    dr = dag.create_dagrun(
        state=DagRunState.RUNNING,
        execution_date=execution_date,
        run_id=run_id,
        start_date=start_date or execution_date,
        session=session,
        conf=conf,  # type: ignore
    )
    return dr



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sql_cli', 'sql_cli.astro', 'sql_cli.operators', 'sql_cli.utils']

package_data = \
{'': ['*'], 'sql_cli': ['macros/*', 'templates/*']}

install_requires = \
['Jinja2>=2.10.1,<=3.1',
 'PyYAML>=5.4.1',
 'apache-airflow>2.1',
 'astro-sdk-python[all]>=1.2.2,<2.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'python-frontmatter>=1.0.0,<2.0.0',
 'typer[all]>=0.6.1,<0.7.0']

extras_require = \
{':python_version < "3.8"': ['networkx>=2.6.3,<3.0.0'],
 ':python_version >= "3.8"': ['networkx>=2.8.7,<3.0.0']}

entry_points = \
{'console_scripts': ['flow = sql_cli.__main__:app']}

setup_kwargs = {
    'name': 'astro-sql-cli',
    'version': '0.2.2a2',
    'description': 'Empower analysts to build workflows to transform data using SQL',
    'long_description': '# Astro SQL CLI\n\nEmpower analysts to build workflows to transform data using SQL.\n\nFind out more in the [docs](https://docs.astronomer.io/astro/cli/sql-cli).\n\n## Getting started\n\n### Install\n\n```bash\n    pip install astro-sql-cli\n```\n\n### Try it out\n\n```bash\n    flow version\n```\n\n\n## Development\n\nSetup your local environment:\n\n```bash\n    make setup\n```\n\nTry your own version of the SQL CLI locally:\n\n```bash\n    poetry run flow version\n```\n\nRun the tests:\n\n```bash\n    make test\n```\n',
    'author': 'Astronomer',
    'author_email': 'humans@astronomer.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
